import os
import gc
import numpy as np
import struct
import mutagen
import pickle
import unicodedata
import json
import sys
import shutil
from tqdm import tqdm
from pydub import AudioSegment
from mutagen.easyid3 import EasyID3
from mutagen.easymp4 import EasyMP4
from pprint import pprint
from colorama import Fore, Back, init


class Coloring(object):
    @staticmethod
    def yellow_text(t):
        return Fore.YELLOW + t + Fore.RESET

    @staticmethod
    def blue_text(t):
        return Fore.BLUE + t + Fore.RESET

    @staticmethod
    def red_text(t):
        return Fore.RED + t + Fore.RESET

    @staticmethod
    def green_text(t):
        return Fore.GREEN + t + Fore.RESET

    @staticmethod
    def yellow_bgcolor(t):
        return Back.YELLOW + t + Back.RESET

    @staticmethod
    def blue_bgcolor(t):
        return Back.BLUE + t + Back.RESET

    @staticmethod
    def red_bgcolor(t):
        return Back.RED + t + Back.RESET

    @staticmethod
    def green_bgcolor(t):
        return Back.GREEN + t + Back.RESET


class Futago(object):
    def __init__(self):
        self.input_dir = "input"
        self.output_dir = "output"
        self.resource_dir = "resource"
        self.bit_rate = 16
        self.sin_wav = 1.0
        self.sampling_rate = 44100
        self.base_sampling_rate = 100
        self.channel = 1
        self.difference = 5
        self.volume = 0
        self.sound_sec_limited = 1800
        self.enable_fadein = False
        self.enable_fadeout = False
        self.fadein_duration = 3500
        self.fadeout_duration = 3500
        self.left_futago_raw_data = f"left_raw_data_{self.base_sampling_rate}Hz"
        self.right_futago_raw_data = f"right_raw_data_{self.base_sampling_rate + self.difference}Hz"
        self.ext = (".mp3", ".aac", ".m4a", ".mp4", ".ogg", ".wav", ".wma", ".flv",)
        self.file_name = f"L{self.base_sampling_rate}-R{self.base_sampling_rate + self.difference}Hz.wav"
        self.futago_path = os.path.join(self.output_dir, self.file_name)
        self.conf_path = "config.json"

        # 双子作成後、ファイルを所定のフォルダに移動させるか？
        self.is_file_move_flag = False
        self.file_moved_dist_path = ""

    def after_file_move(self):
        if self.is_file_move_flag and (self.file_moved_dist_path and os.path.exists(self.file_moved_dist_path)):
            print("ファイルを移動しています...")
            self.file_moved_dist_path = os.path.normpath(self.file_moved_dist_path)
            for file in os.listdir(self.output_dir):
                try:
                    shutil.move(src=os.path.join(self.output_dir, file), dst=self.file_moved_dist_path)
                except (OSError, shutil.Error):
                    print(f"{os.path.relpath(file)}の移動に失敗しました...")
                finally:
                    print(f"{self.output_dir}内のファイルを{self.file_moved_dist_path}へ移動しました。")

    def wave_creator(self, wave=1.0, f0=100, fs=44100, time=1):
        try:
            for n in np.arange(time * fs):
                s = wave * np.sin(2 * np.pi * f0 * n / fs)  # 正弦波の計算
                if s > wave:
                    s = wave
                if s < -wave:
                    s = -wave
                yield struct.pack("h", int(s * 32767.0))
        except MemoryError:
            print("メモリ不足により、プログラムが作成を拒否しています。秒数を短くするか、もっと良いPCに買い換えてください。")
            sys.exit(1)
        finally:
            self.force_gc(wave, f0, fs, time)

    @staticmethod
    def futago_maker(raw_data, t):
        sound = AudioSegment(
            data=raw_data * t,
            sample_width=2,
            frame_rate=44100,
            channels=1,
        )
        return sound

    @staticmethod
    def get_raw_data(d):
        b = b""
        for x in tqdm(d):
            b += x
        return b

    @staticmethod
    def futago_dump(obj, f_name):
        with open(f_name, mode="wb") as fp:
            pickle.dump(obj, fp)

    @staticmethod
    def futago_read(f_name):
        with open(f_name, mode="rb") as fp:
            return pickle.load(fp)

    def init_resource_dumps(self):
        left_raw_data = self.get_raw_data(self.wave_creator(
            wave=self.sin_wav,
            f0=self.base_sampling_rate,
            fs=self.sampling_rate,
            time=1,
        ))
        right_raw_data = self.get_raw_data(self.wave_creator(
            wave=self.sin_wav,
            f0=self.base_sampling_rate + self.difference,
            fs=self.sampling_rate,
            time=1,
        ))

        self.futago_dump(obj=left_raw_data, f_name=os.path.join(self.resource_dir, self.left_futago_raw_data))
        self.futago_dump(obj=right_raw_data, f_name=os.path.join(self.resource_dir, self.right_futago_raw_data))
        print(f"生成に使用するバイナリファイルを{self.resource_dir}フォルダに作成しました...")

    @staticmethod
    def fadein_control(fp, msec=3000):
        return fp.fade_in(duration=msec)

    @staticmethod
    def fadeout_control(fp, msec=3000):
        return fp.fade_out(duration=msec)

    def futago_generator(self, t=10, flag=True):
        left = os.path.join(self.resource_dir, self.left_futago_raw_data)
        right = os.path.join(self.resource_dir, self.right_futago_raw_data)
        if flag:
            if os.path.exists(left) and os.path.exists(right):
                left_audio = self.futago_maker(raw_data=self.futago_read(left), t=t)
                right_audio = self.futago_maker(raw_data=self.futago_read(right), t=t)
                fp = AudioSegment.from_mono_audiosegments(left_audio, right_audio)
                fp = self.volume_change(fp=fp, volume=self.volume)
                if self.enable_fadein:
                    fp = self.fadein_control(fp, self.fadein_duration)
                if self.enable_fadeout:
                    fp = self.fadeout_control(fp, self.fadeout_duration)
                # 書き込み
                self.write(write_name=self.futago_path, audio=fp)
            else:
                self.init_resource_dumps()
                self.futago_generator(t=t, flag=True)
        else:
            if os.path.exists(left) and os.path.exists(right):
                left_audio = self.futago_maker(raw_data=self.futago_read(left), t=t)
                right_audio = self.futago_maker(raw_data=self.futago_read(right), t=t)
                fp = AudioSegment.from_mono_audiosegments(left_audio, right_audio)
                fp = self.volume_change(fp=fp, volume=self.volume)
                if self.enable_fadein:
                    fp = self.fadein_control(fp, self.fadein_duration)
                if self.enable_fadeout:
                    fp = self.fadeout_control(fp, self.fadeout_duration)
                return fp
            else:
                self.init_resource_dumps()
                self.futago_generator(t=t, flag=False)

    @staticmethod
    def get_audio_file(file_path):
        audio_path = os.path.join(file_path)
        ext = os.path.splitext(audio_path)[1]
        if ext == ".wav":
            print(f"{audio_path}を読み込んでいます...")
            return AudioSegment.from_wav(audio_path)
        elif ext == ".mp3":
            print(f"{audio_path}を読み込んでいます...")
            return AudioSegment.from_mp3(audio_path)
        elif ext == ".ogg":
            print(f"{audio_path}を読み込んでいます...")
            return AudioSegment.from_ogg(audio_path)
        elif ext == ".flv":
            print(f"{audio_path}を読み込んでいます...")
            return AudioSegment.from_flv(audio_path)
        else:
            return AudioSegment.from_file(audio_path, ext[1:])

    @staticmethod
    def check_exists_dir(*dirs):
        for dir_name in dirs:
            if not os.path.exists(dir_name):
                os.mkdir(dir_name)
                print(f"{os.path.normpath(os.path.join(os.path.dirname(__file__), dir_name))}フォルダを作成しました")

    @staticmethod
    def over_lay(futago, input_audio):
        return input_audio.overlay(futago, loop=True)

    @staticmethod
    def volume_change(fp, volume=0):
        return fp + volume

    @staticmethod
    def write(audio, write_name=""):
        print(f"{write_name}へと書き込み中です...")
        audio.export(
            write_name,
            bitrate="320k",
            id3v2_version="3",
        )
        print(f"{write_name}を作成しました...")

    @staticmethod
    def force_gc(*obj):
        for i in obj:
            del i
        gc.collect()

    def tag_parser(self, fn):
        audio_ext = os.path.splitext(fn)[1]
        if audio_ext == ".mp4" or audio_ext == ".m4a":
            try:
                t = EasyMP4(fn)
            except mutagen.MutagenError:
                return {}
            t = {k: t[k][0] for k in t.keys()}
            return self.tag_parse_to_str(t)
        elif audio_ext == ".ogg" or audio_ext == ".flac":
            pass
        elif audio_ext == ".mp3":
            try:
                t = dict(EasyID3(fn).items())
            except mutagen.MutagenError:
                return {}
            # listから出してあげる
            t = {k: t[k][0] for k in t.keys()}
            return self.tag_parse_to_str(t)

    @staticmethod
    def save_tags(audio_path, tags):
        audio_ext = os.path.basename(audio_path)
        if audio_ext == ".mp3":
            try:
                audio = EasyID3(audio_path)
            except mutagen.MutagenError:
                return False
            for k, v in tags.items():
                audio[k] = v
            audio.save()
        elif audio_ext == ".mp4" or audio_ext == "m4a":
            try:
                audio = EasyMP4(audio_path)
            except mutagen.MutagenError:
                return False
            for k, v in tags.items():
                audio[k] = v
            audio.save()

    @staticmethod
    def tag_parse_to_str(tags):
        for k, v in tags.items():
            if k == "title":
                tags[k] = os.path.splitext(v)[0]
            elif isinstance(v, int):
                tags[k] = str(v)
        return tags

    @staticmethod
    def delete_on_variable_list(d):
        del d["file_name"]
        del d["futago_path"]
        del d["conf_path"]
        del d["left_futago_raw_data"]
        del d["right_futago_raw_data"]
        return d

    def check_exists_config_file(self):
        with open(self.conf_path, mode="wt", encoding="utf-8") as fp:
            js = json.dumps(self.delete_on_variable_list(self.__dict__.copy()), indent=4)
            fp.write(js)
        print(f"{self.conf_path}を作成しました。")

    def set_config_file_parameter(self):
        with open(self.conf_path, mode="rt", encoding="utf-8") as fp:
            val = json.loads(fp.read())
            self.input_dir = val["input_dir"]
            self.output_dir = val["output_dir"]
            self.resource_dir = val["resource_dir"]
            self.bit_rate = val["bit_rate"]
            self.sin_wav = val["sin_wav"]
            self.sampling_rate = val["sampling_rate"]
            self.base_sampling_rate = val["base_sampling_rate"]
            self.channel = val["channel"]
            self.difference = val["difference"]
            self.volume = val["volume"]
            self.sound_sec_limited = val["sound_sec_limited"]
            self.enable_fadein = val["enable_fadein"]
            self.enable_fadeout = val["enable_fadeout"]
            self.fadein_duration = val["fadein_duration"]
            self.fadeout_duration = val["fadeout_duration"]
            self.ext = val["ext"]
            self.left_futago_raw_data = f"left_raw_data_{self.base_sampling_rate}Hz"
            self.right_futago_raw_data = f"right_raw_data_{self.base_sampling_rate + self.difference}Hz"
            self.file_name = f"L{self.base_sampling_rate}-R{self.base_sampling_rate + self.difference}Hz.wav"
            self.futago_path = os.path.join(self.output_dir, self.file_name)
            self.is_file_move_flag = val["is_file_move_flag"]
            self.file_moved_dist_path = val["file_moved_dist_path"]

    def initialize(self):
        # config.jsonの存在確認、書き込み作成
        if not os.path.exists(self.conf_path):
            self.check_exists_config_file()
        else:
            self.set_config_file_parameter()

        top_text = Coloring.red_text("初期化中...必要なファイルを生成しています☆")
        bottom_text = Coloring.red_text("初期化完了!!!!")
        yellow_fake_star = Coloring.yellow_text("*")
        top_text_count = self.get_star_count(txt=top_text)
        bottom_text_count = self.get_star_count(txt=bottom_text)

        print(f"{yellow_fake_star * top_text_count}{top_text}{yellow_fake_star * top_text_count}")
        self.check_exists_dir(self.input_dir, self.output_dir, self.resource_dir)
        l_path = os.path.join(self.resource_dir, self.left_futago_raw_data)
        r_path = os.path.join(self.resource_dir, self.right_futago_raw_data)
        if not os.path.exists(l_path) or not os.path.exists(r_path):
            self.init_resource_dumps()
        print(f"{yellow_fake_star * bottom_text_count}{bottom_text}{yellow_fake_star * bottom_text_count}\n")

    def pure_futago_maker(self):
        print(f"{self.futago_path}を生成します...")
        notice = f"双子の再生時間を  '半角数字の整数' で入力して下さい。" \
                 f"\n1=1秒なので、60=1分, 300=5分になります。" \
                 f"現在のバージョンでは最大で{self.sound_sec_limited}秒={int(self.sound_sec_limited / 60)}分までとなります。\n\n"
        sec = ""
        err_desc = Coloring.red_text("エラー！！入力は空Enterもしくは半角数値のみとなります。")
        try:
            sec = input(notice)
            assert sec.isdigit(), err_desc
            assert int(sec) <= self.sound_sec_limited, err_desc

        except (AssertionError, ValueError) as e:
            print(e)
            self.pure_futago_maker()
        finally:
            print(f"入力値: {sec}")
        sec = int(sec)
        self.futago_generator(t=sec, flag=True)

    @staticmethod
    def get_strings_count(text):
        count = 0
        for c in text:
            if unicodedata.east_asian_width(c) in 'FWA':
                count += 2
            else:
                count += 1
        return count

    def get_star_count(self, max_length=50, txt=""):
        count = max_length - self.get_strings_count(txt)
        return count // 2

    def to_number(self, num):
        if num == 0 or isinstance(num, int):
            return num
        elif num.isdigit():
            return int(num)
        else:
            try:
                num = int(num)
            except ValueError:
                print(Coloring.red_text("エラー！！半角数値を入力して下さい。"))
                return self.to_number(input())
            finally:
                # print(type(num), f"入力した値={num}")
                pass
            return num

    def permit_fadein(self):
        if not self.enable_fadein:
            print(Coloring.green_text("フェードインエフェクトを有効にしますか？"))
            check_flag = Coloring.yellow_text("許可") if self.enable_fadein else Coloring.red_text("不許可")
            print("現在の設定は" + check_flag + "です。")
            if input("コマンド`y`で許可します。") == "y":
                self.enable_fadein = True
            check_flag = Coloring.yellow_text("許可") if self.enable_fadein else Coloring.red_text("不許可")
            print(check_flag + "しました。")

    def permit_fadeout(self):
        if not self.enable_fadeout:
            print(Coloring.green_text("フェードアウトエフェクトを有効にしますか？"))
            check_flag = Coloring.yellow_text("許可") if self.enable_fadeout else Coloring.red_text("不許可")
            print("現在の設定は" + check_flag + "です。")
            if input("コマンド`y`で許可します。") == "y":
                self.enable_fadeout = True
            check_flag = Coloring.yellow_text("許可") if self.enable_fadeout else Coloring.red_text("不許可")
            print(check_flag + "しました。")

    def ui_volume_change(self, vol=0):
        upper_limit = 100
        lower_limit = -100
        err_desc = f"音量は現在の音量が下限({lower_limit}) <= 現在の音量({self.volume}) <= 上限({upper_limit})になるように入力して下さい。"
        print(f"現在の双子の音量は{vol}です。加算したい値を入力して下さい。\n現在の音量: {self.volume} 上限: {upper_limit} 下限: {lower_limit}")
        user_input = self.to_number(input())
        if user_input == 0 or user_input:
            if lower_limit <= user_input <= upper_limit:
                self.volume = user_input
                print(f"音量の変更に成功しました...。現在の音量: {self.volume}")
                input("Enterを押して下さい...")
                return True
            else:
                print(err_desc)
                self.ui_volume_change(self.volume)
        else:
            print(err_desc)
            self.ui_volume_change(self.volume)

    def menu(self):
        desc_enter = Coloring.yellow_text("空Enter: ")
        desc_1key = Coloring.yellow_text("1: ")
        desc_volume = Coloring.yellow_text("v: ")
        desc_cleaner = Coloring.yellow_text("c: ")
        top_star_text = Coloring.red_text("コマンド一覧")
        yellow_equal = Coloring.yellow_text("=")
        blue_star = Coloring.blue_text("☆")
        yellow_star = Coloring.yellow_text("☆")
        bottom_star_text = Coloring.red_text("コマンド終わり")
        err_desc = Coloring.red_text("エラー！！入力は空Enterもしくは半角数値のみとなります。")
        option_list_text = Coloring.blue_text("オプション機能")
        # コマンド成功値のリスト
        input_assert_list = ("", "1", "v", "c")

        top_star_sum = self.get_star_count(txt=top_star_text)
        bottom_star_sum = self.get_star_count(txt=bottom_star_text)
        option_star_sum = self.get_star_count(txt=option_list_text)

        desc_current_volume = Coloring.green_text(str(self.volume))
        mode_context = f"{desc_enter}{self.input_dir}フォルダ内にある音声ファイルを元に双子の生成を行います。\n" \
                       f"{desc_1key}純粋な双子音声の出力を行います。" \
                       f"最大で{self.sound_sec_limited}秒={int(self.sound_sec_limited / 60)}分の双子が生成出来ます。\n" \
                       f"{yellow_star}{'-' * (option_star_sum - 2)}↓{option_list_text}↓" \
                       f"{'-' * (option_star_sum - 2)}{yellow_star}\n" \
                       f"{desc_cleaner}お掃除機能です。本プログラムが自動生成したフォルダ、ファイルを全部消します。\n" \
                       f"{desc_volume}双子の音量を調整します。現在の設定は{desc_current_volume}です。"

        print(f"{blue_star}{'=' * (top_star_sum-2)}↓{top_star_text}↓{'=' * (top_star_sum-2)}{blue_star}")
        print(mode_context)
        print(f"{blue_star}{yellow_equal*(bottom_star_sum - 2)}↑{bottom_star_text}↑"
              f"{yellow_equal*(bottom_star_sum - 2)}{blue_star}")

        mode = input("モードを入力して下さい...\n")
        if mode in input_assert_list:
            if mode == "":
                self.permit_fadein()
                self.permit_fadeout()
                self.run()
            elif mode == "1":
                self.permit_fadein()
                self.permit_fadeout()
                self.pure_futago_maker()
            elif mode == "v":
                self.ui_volume_change(self.volume)
                self.menu()
            elif mode == "c":
                self.cleaner()
        else:
            print(err_desc)
            self.menu()

    def run(self):
        result = [x for x in os.listdir(self.input_dir) if os.path.splitext(x)[1] in self.ext]

        if result:
            marge_audio = None

            print("以下のファイルに対して処理を行います...")
            print("#" * 50)
            pprint(result)
            print("#" * 50)

            for name in tqdm(result):
                tag = self.tag_parser(fn=os.path.join(self.input_dir, name))
                path = os.path.join(self.output_dir,
                                    os.path.splitext(name)[0] + "(双子可済み)" + os.path.splitext(name)[1])
                if os.path.exists(path):
                    print(f"{path}は既に存在するため、スキップします...")
                    continue
                else:
                    try:
                        marge_audio = self.get_audio_file(os.path.join(self.input_dir, name))
                        futago_audio = self.futago_generator(flag=False, t=round(len(marge_audio) / 1000))
                        self.write(write_name=path, audio=self.over_lay(futago_audio, marge_audio))
                        if os.path.splitext(name)[1] in (".mp3", ".mp4", ".m4a"):
                            self.save_tags(audio_path=path, tags=tag)
                    except MemoryError:
                        print("ファイルが読めませんでした!!\n"
                              "※現在の所、あまり巨大な音声ファイルを読み込むとメモリリークで死にます。")
                    finally:
                        self.force_gc(marge_audio)
        else:
            print(f"{self.input_dir}フォルダ内に合成用の音声ファイルが存在しません...")
            print("上記フォルダ内に音声ファイルを打ち込んでから、本プログラムを再度実行して下さい。")

    def cleaner(self):
        target = ("release", "__pycache__", "build", "input", "output", "resource", ".spec", "config.json")
        print(Coloring.yellow_text("ゴミ掃除を開始します...。") + "本当によろしいですか？^\n対象フォルダ / ファイル一覧")
        pprint(target, indent=4)
        if input("y/n") == "y":
            tar = [x for x in os.listdir(".") if x in target or ".spec" in x]

            for t in tar:
                if os.path.isdir(t):
                    try:
                        os.rmdir(t)
                    except OSError:
                        shutil.rmtree(t)
                    finally:
                        print(f"{Coloring.green_text(t)}フォルダを削除しました。")
                elif os.path.isfile(t):
                    os.remove(t)
                    print(f"{Coloring.green_text(t)}を削除しました。")
            return
        else:
            self.menu()


init()
f = Futago()
f.initialize()
f.menu()
f.after_file_move()
print(Coloring.green_text("すべての処理が完了しました。Enterキーを押すとプログラムを終了します。") + "\n")
input()
sys.exit(0)
